// ==============================================================
//
//  Copyright (c) 1999-2004 by Alex Vinokur.
//
//  For conditions of distribution and use, see
//  copyright notice in version.h
//
// ==============================================================


// ##############################################################
// ==================================
//  n-ary Huffman Template Algorithm
// ==================================
//
//  FILE : raw2.h
//
//  AUTHOR : Alex Vinokur
//
//  DESCRIPTION :
//
//
//         Treatment of raw frequencies (Template Implentation)
//         ----------------------------------------------------
//
// ##############################################################

#ifndef RAW2_H
#define RAW2_H


// =============
#include "raw.h"
// =============



// -------------
// Constructor-0
template <typename WEIGHT, unsigned int ARY>
MixedHuffmanTree <WEIGHT, ARY>::MixedHuffmanTree ()
	:					
	LoadedHuffmanTree<string, WEIGHT, ARY> ()
{
  assert (0);
}


// -------------
// Constructor-1
template <typename WEIGHT, unsigned int ARY>
MixedHuffmanTree <WEIGHT, ARY>::MixedHuffmanTree (
	const vector<string>&	text_file_names_i,
	const vector<string>&	msg_file_name_i,
        size_t			max_ext_freq_stage_i
	)
	: 
	LoadedHuffmanTree<string, WEIGHT, ARY> (),
	text_file_names_ (text_file_names_i),					
        max_ext_freq_stage_ (max_ext_freq_stage_i),
	msg_file_name_ (msg_file_name_i)					
{
const string text_file_name_str ("Text file names");
const string msg_file_name_str ("Message file names");
const string shift1 (5, '#');

size_t setw1 = 0;
size_t setw2 = 0;
  setw1 = MAX_VALUE (setw1, text_file_name_str.size());
  setw1 = MAX_VALUE (setw1, msg_file_name_str.size());

  setw2 = shift1.size() + 1 + setw1 + 3;

  cout << endl;
  cout << string (setw2, '#') << endl;

  cout << shift1 
       << " " 
       << setw (setw1) 
       << text_file_name_str.c_str()
       << " : ";
  copy (text_file_names_.begin(), text_file_names_.end(), ostream_iterator<string> (cout, " "));
  cout << endl;

  cout << shift1 
       << " " 
       << setw (setw1) 
       << msg_file_name_str.c_str()
       << " : ";
  copy (msg_file_name_.begin(), msg_file_name_.end(), ostream_iterator<string> (cout, " "));
  cout << endl;

  cout << string (setw2, '#') << endl;
  cout << endl;


  build_source_strings ();
  compute_extended_frequencies ();

  assert (recomputed_frequencies_.empty());
  for (size_t i = 0; i < ext_frequencies_.size(); i++) 
  {
    recomputed_frequencies_.push_back (false);
  }
  assert (recomputed_frequencies_.size() == ext_frequencies_.size());


  for (size_t i = 1; i <= ext_frequencies_.size(); i++)
  {
    weighted_code_sizes_sum_.push_back (0);

    create_mixed_tree (i);

    if (i > 1)	
    {
      get_superior_compression_factor (i - 1, i);
      get_actual_compression_factor (i - 1, i);
    }


    if (i == 1)	// Temporary ???
    {
      recompute_frequencies (i);
      show_all_frequencies (" (" + to_str(i) + ")" );
    }

  }

}


// -------------
// Destructor
template <typename WEIGHT, unsigned int ARY>
MixedHuffmanTree <WEIGHT, ARY>::~MixedHuffmanTree ()
{
}




// -------------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::create_vect_symbols_codes (size_t alphabet_level_i)
{
  vect_symbols_codes_.clear();
typename map<string, WEIGHT>::iterator iter;

  for (size_t i = 0; 
               (i < alphabet_level_i) && (i < ext_frequencies_.size()); 
               i++
               )
  {
    for (iter = ext_frequencies_[i].begin();
         iter != ext_frequencies_[i].end();
         iter++
         )
    {
      vector<CODE> cur_path;

      if (!encodeOneSymbol (iter->first, cur_path))
      {
        assert (0);
      }

      vect_symbols_codes_.push_back (AuxPairSymbolCode (iter->first, cur_path));
    }
  }
  stable_sort (vect_symbols_codes_.begin(), vect_symbols_codes_.end(), lessAuxPairSymbolCode ());  

}

// -------------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::create_mixed_tree (size_t alphabet_level_i)
{
  assert (alphabet_level_i > 0);

vector<Cell<string, WEIGHT> >	specific_data_vector (fill_data_vector (alphabet_level_i));

  clearBasicHuffmanTree ();
  doBasicHuffmanTree (specific_data_vector);

  create_vect_symbols_codes (alphabet_level_i);

  assert (weighted_code_sizes_sum_.size() == alphabet_level_i);
  weighted_code_sizes_sum_[alphabet_level_i - 1] = getWeightedCodeSizesSum ();


ostringstream oss;
  copy (
        text_file_names_.begin(),
        text_file_names_.end(),
        ostream_iterator<string> (oss, " ")
        );


  showAll ("[ Alphabet level = " 
          + to_str (alphabet_level_i) 
          + " ] " 
          + string ("Creating Mixed Huffman Tree for text files : ") 
          + oss.str()
          );


   

  for (count_t i = 0; i < msg_file_name_.size(); i++)
  {
    cout << endl;
    cout << endl;
    cout << string (12, ' ')
         << string (5, '#')
         << " "
         << "Message-"
         << (i + 1)
         << " "
         << string (5, '#')
         << endl;
    cout << endl;

    cerr << "---> [" 
         << alphabet_level_i 
         << "-Alphabet" 
         << "] Message-" 
         << (i + 1) << " : " 
         << msg_file_name_[i] 
         << endl;


const string source_msg_str (build_string_from_file (msg_file_name_[i]));

    if (source_msg_str.empty())
    {
      FATAL_MSG ("Message file <"
              << msg_file_name_[i]
              << "> is empty"
              << endl
              << FATAL_SHIFT
              << "The file must contain message to be encoded"
              );
      continue;
    }

    cout << "Source Message  : file  " << msg_file_name_[i] << endl; 

    encode_decode_message (
		false,
		msg_file_name_[i], 
		source_msg_str, 
		alphabet_level_i
		);

    cout << endl;
    cout << endl;


  }


  for (count_t i = 0; i < text_file_names_.size(); i++)
  {
    cout << endl;
    cout << endl;
    cout << string (12, ' ')
         << string (5, '#')
         << " "
         << "Source-Text-"
         << (i + 1)
         << " "
         << string (5, '#')
         << endl;
    cout << endl;


    cerr << "---> [" 
         << alphabet_level_i 
         << "-Alphabet" 
         << "] Source-" 
         << (i + 1) << "  : " 
         << text_file_names_[i] 
         << endl;

const string source_msg_str (build_string_from_file (text_file_names_[i]));

    if (source_msg_str.empty())
    {
      FATAL_MSG ("Source text file <"
              << text_file_names_[i]
              << "> is empty"
              << endl
              << FATAL_SHIFT
              << "The file must contain text to be encoded"
              );
      continue;
    }

    cout << "Source Text  : file  " << text_file_names_[i] << endl; 

    encode_decode_message (
		true,
		text_file_names_[i], 
		source_msg_str, 
		alphabet_level_i
		);

    cout << endl;
    cout << endl;


  }


}


// -------------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::build_source_strings ()
{
  for (size_t i = 0; i < text_file_names_.size(); i++)
  {
    source_strings_.push_back (build_string_from_file (text_file_names_[i]));
  }
}


// ---------
template <typename WEIGHT, unsigned int ARY>
size_t MixedHuffmanTree <WEIGHT, ARY>::get_pattern_frequencies_s (const string& str_i, const string& pattern_i)
{
size_t ret_value = 0;
size_t pos = 0;
const size_t pattern_size = pattern_i.size();
const size_t str_size = str_i.size();
  if (str_size < pattern_size)	return 0;

  while ((pos + pattern_size) <= str_size) 
  {
    pos = str_i.find (pattern_i, pos);
    if (pos == string::npos) break;
    ret_value++;
    pos += pattern_size;
  }

  return ret_value;
}


// ---------
template <typename WEIGHT, unsigned int ARY>
vector<string> MixedHuffmanTree <WEIGHT, ARY>::remove_substr_s (const string& str_i, const string& substr_i)
{
vector<string>	ret_vect;
size_t		sub_size = substr_i.size(); 
size_t	pos1 = 0;	
size_t	pos2;	

  while ((pos2 = str_i.find (substr_i, pos1)) != string::npos)
  {
    assert (pos2 >= pos1);

    if (pos2 > pos1) ret_vect.push_back (str_i.substr (pos1, pos2 - pos1));

    pos1 = pos2 + sub_size;
  }
  if (pos1 < str_i.size()) ret_vect.push_back (str_i.substr (pos1));

  return ret_vect;
}

// ---------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::show_frequencies_s (
		const map<string, WEIGHT>& map_i, 
		const string&	cur_source_str_i,
		size_t		pattern_size_i,
		const string&	info_type_i,
		const string&	msg_i
		)
{
typename map<string, WEIGHT>::const_iterator iter;

  if (!msg_i.empty())
  {
    cout << "\t" << string (9, '#') << " " << msg_i << " " << string (9, '#') << endl;
  }

  if (!info_type_i.empty())
  {
    cout << "\t### Map of " << info_type_i << " frequencies (pattern size = " << pattern_size_i << ") ###" << endl;
  }

  if (!cur_source_str_i.empty())
  {						  
    cout << "\t### Current source string : " << gstr_symbol (cur_source_str_i) << endl;
  }

vector<size_t>	tmp_sizes;
  for (iter = map_i.begin(); iter != map_i.end(); iter++)
  {
    tmp_sizes.push_back (gstr_symbol (iter->first).size());
  }

size_t max_setw = *(max_element (tmp_sizes.begin(), tmp_sizes.end()));
size_t count = 0;
  for (iter = map_i.begin(); iter != map_i.end(); iter++)
  {
    cout << ""
         << setw (3)
         << setfill (' ')
         << right
         << pattern_size_i
         << "."
         << setw (5)
         << setfill (' ')
         << right
         << (++count)
         << " : "
         << setw (max_setw + 5)
         << setfill (' ')
         << left
         << gstr_symbol (iter->first).c_str()
         << setw (10)
         << setfill (' ')
         << right
         << iter->second
         << endl;
  }
  cout << endl;
}


// ---------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::show_frequencies () const
{

ostringstream oss;
  copy (
        text_file_names_.begin(),
        text_file_names_.end(),
        ostream_iterator<string> (oss, " ")
        );


  cout << "   #########################" << endl;
  cout << "   ### Mixed frequencies ###" << endl;
  cout << "   === Files : " << oss.str() << endl;
  cout << "   === Source strings : " << endl;
  for (size_t i = 0; i < source_strings_.size(); i++)
  {
    cout << "       " << gstr_symbol (source_strings_[i]) << endl;
  }
  cout << "   #########################" << endl;

  for (size_t i = 0; i < ext_frequencies_.size(); i++)
  {
   cout << endl;
   cout << "      --- Pattern size = " << (i + 1) << " ---" << endl;
   show_frequencies_s (
	ext_frequencies_[i], 
	string(), 
	i + 1, 
	string(), 
	string() 
	);
  }
}


// ---------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::init_frequencies_s (
		map<string, WEIGHT>& map_o,
		size_t pattern_size_i, 
		const string& str_i
		)
{
  assert (pattern_size_i > 0);

  assert (str_i.size() >= pattern_size_i);

const size_t the_size = str_i.size() - pattern_size_i + 1;
  for (size_t i = 0; i < the_size; i++)
  {
    const string the_str (str_i.substr (i, pattern_size_i));
    if (map_o.count (the_str) == 0)
    {
      map_o[the_str] = 0;
    }
  } 
}
		 

// ---------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::compute_frequencies_s (
		map<string, WEIGHT>&	map_io, 
		const string&		str_i 
		)
{
typename map<string, WEIGHT>::iterator iter;
  for (iter = map_io.begin(); iter != map_io.end(); iter++)
  {
    iter->second = get_pattern_frequencies_s (str_i, iter->first);
  }

}


// ---------
template <typename WEIGHT, unsigned int ARY>
typename map<string, WEIGHT>::const_iterator MixedHuffmanTree <WEIGHT, ARY>::get_max_frequency_s (const map<string, WEIGHT>& map_i)
{
size_t	max_freq = 0;
typename map<string, WEIGHT>::const_iterator iter;
typename map<string, WEIGHT>::const_iterator ret_iter = map_i.end();

  assert (!map_i.empty());
  for (iter = map_i.begin(); iter != map_i.end(); iter++)
  {
    if (iter->second > max_freq)
    {
      max_freq = iter->second;
      ret_iter = iter;
    }
  }
  return ret_iter;
}



// ---------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::add_frequencies_s (const map<string, WEIGHT>& source_map_i, map<string, WEIGHT>& result_map_io)
{
typename map<string, WEIGHT>::const_iterator iter = get_max_frequency_s (source_map_i);
  assert (!(iter == source_map_i.end()));

  result_map_io [iter->first] += iter->second; 

}


// ---------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::compute_extended_frequencies (size_t pattern_size_i)
{

map<string, WEIGHT>	source_freqs_map;

  ext_frequencies_.resize (ext_frequencies_.size() + 1);
  accum_sizes_.push_back (0);

  assert (ext_frequencies_.size() == pattern_size_i);

  for (size_t s = 0; s < source_strings_.size(); s++)
  {
    const string source_str (source_strings_[s]);
    if (source_str.size() < pattern_size_i) continue;

    assert (!source_str.empty());
  
    vector<string> str_vect;
  
    str_vect.push_back (source_str);
  
    assert (pattern_size_i > 0);
  
    init_frequencies_s (source_freqs_map, pattern_size_i, source_str);
  
    for (size_t i = 0; !str_vect.empty(); i++)
    {
      string cur_string (*str_vect.rbegin());
      assert (cur_string.size() >= pattern_size_i);
  
      compute_frequencies_s (source_freqs_map, cur_string);
      add_frequencies_s (source_freqs_map, ext_frequencies_[pattern_size_i - 1]);
  
      /*
      cout << endl << endl << endl << endl;
      show_frequencies_s (source_freqs_map, cur_string, pattern_size_i, "preliminary", "Pre-Step-" + to_string (i + 1));
      cout << endl << endl << endl;
      show_frequencies_s (ext_frequencies_[pattern_size_i - 1], cur_string, pattern_size_i, "inter-resulting", "Res-Step-" + to_string (i + 1));
      */
  
      str_vect.erase (str_vect.end() - 1);
  

  
      vector<string> tmp_vect (remove_substr_s (cur_string, get_max_frequency_s (source_freqs_map)->first));
  
      for (size_t j = 0; j < tmp_vect.size(); j++)
      {

        if (tmp_vect[j].size() >= pattern_size_i) str_vect.push_back (tmp_vect[j]);
      }
     
    }
  } // for (size_t s = 0; s < source_strings_.size(); s++)

  if ((*ext_frequencies_.rbegin()).empty())
  {
    ext_frequencies_.erase (ext_frequencies_.end() - 1);
  }

  //show_frequencies_s (ext_frequencies_[pattern_size_i - 1], string(), pattern_size_i, "summary-resulting", "Summary");
}


// ---------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::compute_extended_frequencies ()
{
bool criterion = !source_strings_.empty();
ullong cur_max_freq;
const string pref_note ("Getting frequencies    ");

  cerr << endl << pref_note << " : " << "Start" << endl;
  for (size_t i = 1; criterion; i++)
  {
    assert (i <= (ext_frequencies_.size() + 1));

    compute_extended_frequencies (i);

    cur_max_freq = get_max_frequency_s (ext_frequencies_[i - 1 ])->second;

    cerr << pref_note << " : " << "Stage#" << i << " (";
    switch (i)
    {
      case 0 :
       assert (0);
       break;

      case 1 :
        cerr << "characters";
        break;

      case 2 :
        cerr << "bigrams   ";
        break;

      default:
        cerr << i << "-grams   ";
        break;

    }
    cerr << ") done;  max frequency = " << cur_max_freq << endl;

    assert (i <= ext_frequencies_.size());

    criterion = (
                 (cur_max_freq > 1)
                 &&
                 ((i < max_ext_freq_stage_) || (max_ext_freq_stage_ == 0))
                );
  }

  show_frequencies ();
  cerr << pref_note << " : " << "Finish" << endl << endl;
}



// ---------
template <typename WEIGHT, unsigned int ARY>
vector<Cell<string, WEIGHT> > MixedHuffmanTree <WEIGHT, ARY>::fill_data_vector (size_t alphabet_level_i) const
{
vector<Cell<string, WEIGHT> >	ret_data_vector;

typename map<string, WEIGHT>::const_iterator iter;


  for (size_t i = 0; 
               (i < alphabet_level_i) && (i < ext_frequencies_.size()); 
               i++
               )
  {
    for (iter = ext_frequencies_[i].begin(); 
         !(iter == ext_frequencies_[i].end());
         iter++
        )
    {
      ret_data_vector.push_back (Cell<string, WEIGHT> (iter->first, iter->second));
    }
  }

  return ret_data_vector;

}



// ---------
template <typename WEIGHT, unsigned int ARY>
bool MixedHuffmanTree <WEIGHT, ARY>::encode_decode_message (
	bool  source_text_flag_i,
	const string& msg_file_name_i,
	const string& msg_str_i,
	size_t	alphabet_level_i	
	) const
{
  assert (alphabet_level_i > 0);
  assert (alphabet_level_i <= ext_frequencies_.size());

bool	tmp_bool;
bool	result_enc;
bool	result_dcd = false;
ullong	filesize;

size_t pos = msg_file_name_i.find_last_of (".");
const string head_name (msg_file_name_i.substr (0, pos));

ofstream fs_encoded;
ofstream fs_detailed;
ofstream fs_decoded;


size_t	reduce_counter = 0;
  for (size_t i = 0; i < recomputed_frequencies_.size(); i++)
  {
    reduce_counter += static_cast<size_t>(recomputed_frequencies_[i]);  
  }

ostringstream oss2;
  oss2 << ""
       << "[ " 
       << msg_file_name_i
       << ", "
       << alphabet_level_i 
       << "-Alphabet" 
       //<< "; " 
       //<< setw (10)
       //<< left
       //<< (reduce_counter ? string ("REDUCED-" + to_str (reduce_counter, 2, '0')).c_str() : "RAW") 
       << " ]  ";



//const string ext_head_name (head_name + "-" + to_str (alphabet_level_i) + "-" + to_str (reduce_counter, 2, '0'));
const string ext_head_name (head_name + "-" + to_str (alphabet_level_i));


// -------------
const string encode_tail ("enc");
const string detailed_tail ("exi");
const string decode_tail ("dcd");

const string encoded_file_name (ext_head_name + "." + encode_tail);
const string detailed_file_name (ext_head_name + "." + detailed_tail);
const string decoded_file_name (ext_head_name + "." + decode_tail);


  // ---------
  fs_encoded.open (encoded_file_name.c_str());
  ASSERT (fs_encoded);

  fs_detailed.open (detailed_file_name.c_str());
  ASSERT (fs_detailed);

  fs_decoded.open (decoded_file_name.c_str());
  ASSERT (fs_decoded);


const vector<string> source_msg_vect (build4_vect_msg (msg_str_i, alphabet_level_i));
//const vector<string> source_msg_vect (build5_vect_msg (msg_str_i));


vector<string> decoded_msg_vect;
vector<CODE>   encoded_msg_vect;


  cout << endl;
  if (alphabet_level_i == 1)
  {
    cout << string (7, '-') 
         << " " 
         << "Alphabet contains ordinary symbols (of size 1)" 
         << " " 
         << string (7, '-') 
         << endl;
  }
  else
  {
    cout << string (7, '-') 
         << " " 
         << "Alphabet contains mixed symbols of size 1-" 
         << alphabet_level_i 
         << " " 
         << string (7, '-') 
         << endl;
  }



  copy (
          source_msg_vect.begin(),
          source_msg_vect.end() ,
          ostream_iterator<string> (fs_detailed, "\n")
          );
  fs_detailed.flush();

ostringstream oss;
  result_enc = encodeMsg (source_msg_vect, encoded_msg_vect);
  if (result_enc)
  {
    copy (
          encoded_msg_vect.begin(),
          encoded_msg_vect.end() ,
          ostream_iterator<CODE> (fs_encoded, "")
          );
    copy (
          encoded_msg_vect.begin(),
          encoded_msg_vect.end() ,
          ostream_iterator<CODE> (oss, "")
          );
  }
  else
  {
    cout << "Cannot encode Message : "
         << endl
         << "<"
         << gstr_vector (source_msg_vect) 
         << ">"
         << endl;

    fs_encoded.close(); assert (!fs_encoded.is_open());
    fs_detailed.close(); assert (!fs_detailed.is_open());
    fs_decoded.close();	assert (!fs_decoded.is_open());

    return false;
  }

  result_dcd = decodeMsg (encoded_msg_vect, decoded_msg_vect);
  if (result_dcd)
  {
    copy (
          decoded_msg_vect.begin(),
          decoded_msg_vect.end() ,
          ostream_iterator<string> (fs_decoded, "")
          );
  }
  else
  {
    cout << "Cannot decode encoded Message : "
         << endl
         << "<" 
         << gstr_vector (encoded_msg_vect) 
         << ">" 
         << endl;

    fs_encoded.close(); assert (!fs_encoded.is_open());
    fs_decoded.close();	assert (!fs_decoded.is_open());

    return false;
  }


  // ---------
  assert (fs_encoded.is_open());
  fs_encoded.close();
  assert (!fs_encoded.is_open());


  assert (fs_decoded.is_open());
  fs_decoded.close();
  assert (!fs_decoded.is_open());


  // ----------------------
  assert (result_enc);
  assert (result_dcd);

  assert (alphabet_level_i <= recomputed_frequencies_.size());


  tmp_bool = get_file_size (encoded_file_name, filesize);
  assert (tmp_bool);
  cout << oss2.str() << "Encoded Message : file  " << encoded_file_name << " ( size = " << filesize << " )" << endl; 
  assert (filesize == oss.str().size());

  if (source_text_flag_i)
  {
    assert ((alphabet_level_i - 1) < accum_sizes_.size());
    accum_sizes_ [alphabet_level_i - 1] += filesize;
  }
  cout << oss2.str() << "Decoded Message : file  " << decoded_file_name << endl; 

  assert (build_string_from_file (msg_file_name_i) == build_string_from_file (decoded_file_name));

  return true;
}



// ---------
template <typename WEIGHT, unsigned int ARY>
string MixedHuffmanTree <WEIGHT, ARY>::build_string_from_file (const string& file_name_i) const
{
string	ret_string;
ifstream fin (file_name_i.c_str ());

  // ------------------------------------
  if (!fin)
  {
    FATAL_MSG ("Cannot open file "
              << file_name_i
              << " for reading"
              << endl
              );
    return string();
  }

char ch;
  while ( fin.get (ch) ) ret_string += ch;

  if (ret_string.empty())
  {
    FATAL_MSG ("File "
              << file_name_i
              << " is empty"
              << endl
              );
  }

  fin.close();
  ASSERT (!fin.is_open());

  return ret_string;

}



// ---------
template <typename WEIGHT, unsigned int ARY>
vector<string> MixedHuffmanTree <WEIGHT, ARY>::build1_vect_msg (
		const string& str_msg_i, 
		size_t alphabet_level_i
		) const
{
vector<string>	ret_vect;

string	head_str;  
string	tail_str (str_msg_i);

  while (!tail_str.empty())
  {
    bool ret_bool = get1_msg_head_and_tail (alphabet_level_i, head_str, tail_str);

    if (!ret_bool)
    {
      FATAL_MSG ("Message"
             << endl
             << string (10, '-')
             << endl
             << str_msg_i
             << string (10, '-')
             << endl
             << "contains illegal symbol"
             );
    }

    assert (ret_bool);
    ret_vect.push_back (head_str);
  }  

  return ret_vect;

}



// ---------
template <typename WEIGHT, unsigned int ARY>
vector<string> MixedHuffmanTree <WEIGHT, ARY>::build4_vect_msg (
		const string& str_msg_i, 
		size_t alphabet_level_i
		) const
{
vector<string>	ret_vect;

string	head_str;  
string	tail_str (str_msg_i);

  while (!tail_str.empty())
  {
    bool ret_bool = get4_msg_head_and_tail (alphabet_level_i, head_str, tail_str);

    if (!ret_bool)
    {
      FATAL_MSG ("Message"
             << endl
             << string (10, '-')
             << endl
             << str_msg_i
             << string (10, '-')
             << endl
             << "contains illegal symbol"
             );
    }

    assert (ret_bool);
    ret_vect.push_back (head_str);
  }  

  return ret_vect;

}


// ---------
template <typename WEIGHT, unsigned int ARY>
vector<string> MixedHuffmanTree <WEIGHT, ARY>::build5_vect_msg (
		const string& str_i, 
		size_t alphabet_level_i
		) const
{
vector<string>	ret_vect;
count_t		pos;
string		cur_huf_symbol;
map<ullong, string>	tmp_map1;
string		treated_str (str_i);
vector<ullong>	treated_vect (str_i.size());
const char	no_char = '\0';
const string	no_find_str (1, no_char);

  assert (alphabet_level_i == 0);

  assert (treated_str.find_first_of (no_find_str) == string::npos);  

  for (count_t i = 0; i < treated_vect.size(); i++) treated_vect[i] = i;

  assert (!vect_symbols_codes_.empty());
  for (size_t i = 0; i < vect_symbols_codes_.size(); i++)
  {
    cur_huf_symbol = vect_symbols_codes_[i].ext_symbol_;

    pos = 0;
    while ((pos = treated_str.find (cur_huf_symbol, pos)) != string::npos)
    {
      assert (treated_str.size() == treated_vect.size());
      assert (tmp_map1.count(treated_vect[pos]) == 0);
      tmp_map1 [treated_vect[pos]] = cur_huf_symbol;

      for (count_t k = pos; k < (pos + cur_huf_symbol.size()); k++)
      {
        treated_str[k] = no_char;
      }
    }

  }

map<ullong, string>::iterator	map_iter;

  for (map_iter = tmp_map1.begin(); map_iter != tmp_map1.end(); map_iter++)
  {
    ret_vect.push_back (map_iter->second);
  }

  return ret_vect;

}


// ---------
template <typename WEIGHT, unsigned int ARY>
vector<string> MixedHuffmanTree <WEIGHT, ARY>::build3_vect_msg (
		const string& str_i, 
		size_t alphabet_level_i
		) const
{

  assert (alphabet_level_i == 0);

vector<string>	ret_vect;
  ret_vect.push_back (str_i);

vector<string> tmp2_vect;
bool criterion = true;

  do
  {
    vector<string> tmp_vect (ret_vect);
    ret_vect.clear();

    size_t i;
    for (i = 0; i < tmp_vect.size(); i++)
    {
      tmp2_vect = build3i_vect_msg (tmp_vect[i]);
      assert (!tmp2_vect.empty());
      criterion = (tmp2_vect.size() > 1);
      if (criterion) break;
    }

    copy (tmp_vect.begin(), tmp_vect.begin() + i, back_inserter (ret_vect));
    if (criterion)
    {
      copy (tmp2_vect.begin(), tmp2_vect.end(), back_inserter (ret_vect));
    }
    copy (tmp_vect.begin() + (i + 1), tmp_vect.end(), back_inserter (ret_vect));

    assert (!tmp2_vect.empty());

  } while (criterion);

  return ret_vect;

}

// ---------
template <typename WEIGHT, unsigned int ARY>
vector<string> MixedHuffmanTree <WEIGHT, ARY>::build3i_vect_msg (
		const string& str_i 
		) const
{
vector<string>	ret_vect;
size_t	pos;
bool	check_flag = false;

  assert (!vect_symbols_codes_.empty());
  for (size_t i = 0; i < vect_symbols_codes_.size(); i++)
  {
    pos = str_i.find (vect_symbols_codes_[i].ext_symbol_);
    if (pos == string::npos) continue;

    check_flag = true;

    size_t ssize = vect_symbols_codes_[i].ext_symbol_.size();

    if (pos > 0)	ret_vect.push_back (str_i.substr(0, pos));
    ret_vect.push_back (str_i.substr(pos, ssize));
    if ((pos + ssize) < str_i.size())	ret_vect.push_back (str_i.substr(pos + ssize));

    break;
  }
  assert (check_flag);

  assert (!ret_vect.empty());
  assert (ret_vect.size() >= 1);
  assert (ret_vect.size() <= 3);
  if (ret_vect.size() == 1) assert (ret_vect[0] == str_i);

  return ret_vect;

}

// ---------
template <typename WEIGHT, unsigned int ARY>
bool MixedHuffmanTree <WEIGHT, ARY>::get1_msg_head_and_tail (
			size_t		alphabet_level_i,
			string&		head_o,
			string&		tail_io
			) const
{
typename map<string, WEIGHT>::const_iterator iter;

  assert (alphabet_level_i > 0);
  assert (alphabet_level_i <= ext_frequencies_.size());

size_t item_size;
string tmp_str;

  for (count_t i = 0; i < alphabet_level_i; i++)
  {
    const count_t revrs_i = alphabet_level_i - 1 - i; 
    for (iter = ext_frequencies_[revrs_i].begin(); 
         !(iter == ext_frequencies_[revrs_i].end());
         iter++
        )
    {
      item_size = iter->first.size();
      ASSERT (item_size == (revrs_i + 1));
      assert (!iter->first.empty());

      tmp_str.clear();
      tmp_str = tail_io.substr (0, item_size);
      if (!(tmp_str == iter->first)) continue;

      head_o  = tail_io.substr (0, item_size);     
      tail_io = tail_io.substr (item_size);
      return true;     
    }
  }


  ERROR_MSG ("Message "
             << "contains illegal symbol <"
             << tmp_str
             << ">"
             );
  
  return false;

}



// ---------
template <typename WEIGHT, unsigned int ARY>
bool MixedHuffmanTree <WEIGHT, ARY>::get4_msg_head_and_tail (
			size_t		alphabet_level_i,
			string&		head_o,
			string&		tail_io
			) const
{
  assert (alphabet_level_i > 0);
  assert (alphabet_level_i <= ext_frequencies_.size());

size_t item_size;
string tmp_str;

  for (count_t i = 0; i < vect_symbols_codes_.size(); i++)
  {
    item_size = vect_symbols_codes_[i].ext_symbol_.size();

    tmp_str.clear();
    tmp_str = tail_io.substr (0, item_size);

    if (!(tmp_str == vect_symbols_codes_[i].ext_symbol_)) continue;

    head_o  = tail_io.substr (0, item_size);     
    tail_io = tail_io.substr (item_size);
    return true;     

  }


  ERROR_MSG ("Message "
             << "contains illegal symbol <"
             << tmp_str
             << ">"
             );
  
  return false;

}



// ---------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::recompute_frequencies (size_t alphabet_level_i)
{
  //show_all_frequencies ("Recomputing frequencies : Before step-" + to_str(alphabet_level_i));
  assert (alphabet_level_i <= 1); // Temporary ???

  // XXX assert (alphabet_level_i < recomputed_frequencies_.size());
  assert (!recomputed_frequencies_[alphabet_level_i - 1]);

typename map<string, WEIGHT>::iterator iter; 
typename map<string, WEIGHT>::iterator find_iter; 

  if (alphabet_level_i < recomputed_frequencies_.size())
  {
    for (iter = ext_frequencies_[alphabet_level_i].begin(); 
         !(iter == ext_frequencies_[alphabet_level_i].end());
         iter++
        )
    {
      const string tmp1_string (iter->first.substr (0, iter->first.size() - 1));
      const string tmp2_string (iter->first.substr (1));
  
      find_iter = ext_frequencies_[alphabet_level_i - 1].find (tmp1_string);
      if (find_iter != ext_frequencies_[alphabet_level_i - 1].end())
      {
        assert (iter->second <= find_iter->second);
        find_iter->second -= iter->second;  
      }
  
      find_iter = ext_frequencies_[alphabet_level_i - 1].find (tmp2_string);
      if (find_iter != ext_frequencies_[alphabet_level_i - 1].end())
      {
        assert (iter->second <= find_iter->second);
        find_iter->second -= iter->second;  
      }
  
    } // for (iter = ext_frequencies_[alphabet_level_i].begin()
  } // if (alphabet_level_i < recomputed_frequencies_.size())

  assert ((alphabet_level_i - 1) < recomputed_frequencies_.size());
  recomputed_frequencies_[alphabet_level_i - 1] = true;

const string pref_note ("Recomputing frequencies");

    cerr << endl << pref_note << " : " << "Stage#" << alphabet_level_i << " (";
    switch (alphabet_level_i)
    {
      case 0 :
       assert (0);
       break;

      case 1 :
        cerr << "characters";
        break;

      case 2 :
        cerr << "bigrams   ";
        break;

      default:
        cerr << alphabet_level_i << "-grams   ";
        break;

    }
    cerr << ") done" << endl << endl;;

  //show_all_frequencies ("Recomputing frequencies : After  step-" + to_str(alphabet_level_i));

}


// ---------
template <typename WEIGHT, unsigned int ARY>
void MixedHuffmanTree <WEIGHT, ARY>::show_all_frequencies (const string& msg_i) const
{
  if (!msg_i.empty ()) 
  {
    cout << endl 
         << endl 
         << "\t###### "
         << "Recomputed frequencies"  
         << " : "
         << msg_i 
         << endl;
  }

size_t setw1 = 0;

typename map<string, WEIGHT>::const_iterator iter; 

  // -----------------------------------
  for (size_t i = 0; i < ext_frequencies_.size(); i++)
  {
    for (iter = ext_frequencies_[i].begin();
         iter != ext_frequencies_[i].end();
         iter++
         )
    {
      setw1 = MAX_VALUE (setw1, gstr_symbol (iter->first).size());
    }        
    
  }

  setw1++;
  // -----------------------------------
  for (size_t i = 0; i < ext_frequencies_.size(); i++)
  {
    cout << "\n\t=== " << (i + 1) << "-Alphabet ===" << endl;
    for (iter = ext_frequencies_[i].begin();
         iter != ext_frequencies_[i].end();
         iter++
         )
    {
      cout << setw (setw1) 
           << left 
           << gstr_symbol (iter->first).c_str() 
           << " :::  " 
           << iter->second 
           << endl;
    }        
    
  }

}



// ---------
// ---------
#define tmp_str_a	string ("Huffman tree with ")
#define tmp_str_b1 	string ("-Alphabet")
#define	setw_size1	28


// ---------
template <typename WEIGHT, unsigned int ARY>
double MixedHuffmanTree <WEIGHT, ARY>::get_superior_compression_factor (size_t alphabet1_i, size_t alphabet2_i) const
{
  assert (alphabet1_i < alphabet2_i);
  assert ((alphabet2_i - 1) < weighted_code_sizes_sum_.size());

  assert (weighted_code_sizes_sum_[alphabet1_i - 1] > 0);
  assert (weighted_code_sizes_sum_[alphabet2_i - 1] > 0);

freq_t	wcodesize1 = weighted_code_sizes_sum_[alphabet1_i - 1];
freq_t	wcodesize2 = weighted_code_sizes_sum_[alphabet2_i - 1];


double	ret_superior_factor = static_cast<double>(wcodesize1)/static_cast<double>(wcodesize2);


const string tmp_str_b2 (" : weighted code sizes sum = ");
const string tmp_str_b (tmp_str_b1 + tmp_str_b2);
const string tmp_str_c ("Superior Compression Factor");


ostringstream oss;
  oss  << endl
       << setw (setw_size1)
       << left
       << tmp_str_c
       << " : " 
       << alphabet2_i 
       << "-Alphabet"
       << "/"
       << alphabet1_i
       << "-Alphabet"
       << " = "
       << ret_superior_factor
       << "   ( "
       << wcodesize1
       << "/"
       << wcodesize2
       << " )"
       << endl;


  cerr << oss.str();
			    
size_t size1 = 0;
  size1 = MAX_VALUE (size1, tmp_str_a.size() + tmp_str_b1.size() + 1);
  assert (size1 == setw_size1);
  size1 = MAX_VALUE (size1, oss.str().size());

  cout << endl;
  cout << endl;
  cout << endl;
  cout << string (size1, '-') << endl;
  cout << string ((size1 - tmp_str_c.size())/2 - 1, '-')
       << " "
       << tmp_str_c
       << " "
       << string ((size1 - tmp_str_c.size())/2 - 1, '-')
       << endl;
  cout << tmp_str_a << alphabet1_i << tmp_str_b << wcodesize1 << endl;
  cout << tmp_str_a << alphabet2_i << tmp_str_b << wcodesize2 << endl;
  cout << oss.str();
  cout << string (size1, '-') << endl;

  return ret_superior_factor;
}



// ---------
template <typename WEIGHT, unsigned int ARY>
double MixedHuffmanTree <WEIGHT, ARY>::get_actual_compression_factor (size_t alphabet1_i, size_t alphabet2_i) const
{
  assert (alphabet1_i < alphabet2_i);
  assert ((alphabet2_i - 1) < accum_sizes_.size());

  assert (accum_sizes_[alphabet1_i - 1] > 0);
  assert (accum_sizes_[alphabet2_i - 1] > 0);


ullong	accumsize1 = accum_sizes_[alphabet1_i - 1];
ullong	accumsize2 = accum_sizes_[alphabet2_i - 1];


double	ret_actual_factor = static_cast<double>(accumsize1)/static_cast<double>(accumsize2);


const string tmp_str_b2 (" : encoded sources sizes sum = ");
const string tmp_str_b (tmp_str_b1 + tmp_str_b2);
const string tmp_str_d ("Actual Compression Factor");


ostringstream oss;
  oss  << endl
       << setw (setw_size1)
       << left
       << tmp_str_d
       << " : " 
       << alphabet2_i 
       << "-Alphabet"
       << "/"
       << alphabet1_i
       << "-Alphabet"
       << " = "
       << ret_actual_factor
       << "   ( "
       << accumsize1
       << "/"
       << accumsize2
       << " )"
       << endl;


  cerr << oss.str();


size_t size1 = 0;
  size1 = MAX_VALUE (size1, tmp_str_a.size() + tmp_str_b1.size() + 1);
  assert (size1 == setw_size1);
  size1 = MAX_VALUE (size1, oss.str().size());

  cout << endl;
  cout << endl;
  cout << endl;
  cout << string (size1, '-') << endl;
  cout << string ((size1 - tmp_str_d.size())/2 - 1, '-')
       << " "
       << tmp_str_d
       << " "
       << string ((size1 - tmp_str_d.size())/2 - 1, '-')
       << endl;
  cout << tmp_str_a << alphabet1_i << tmp_str_b << accumsize1 << endl;
  cout << tmp_str_a << alphabet2_i << tmp_str_b << accumsize2 << endl;
  cout << oss.str();
  cout << string (size1, '-') << endl;

  return ret_actual_factor;
}


#endif	// RAW2_H

